const std = @import("std");
const sdl = @import("sdl2");
const vk = @import("vulkan");
const builtin = @import("builtin");
const shaders = @import("shaders");

const Utilities = @import("utilities.zig");
const QueueFamilyIndices = Utilities.QueueFamilyIndices;
const SwapchainDetails = Utilities.SwapchainDetails;
const SwapchainImage = Utilities.SwapchainImage;

const enable_validation_layers = builtin.mode == .Debug;
const validation_layers = [_][*:0]const u8{"VK_LAYER_KHRONOS_validation"};

const apis: []const vk.ApiInfo = &.{
    vk.features.version_1_0,
    vk.features.version_1_1,
    vk.features.version_1_2,
    vk.features.version_1_3,
    vk.extensions.khr_surface,
    vk.extensions.khr_swapchain,
    vk.extensions.ext_debug_utils,
};

const BaseDispatch = vk.BaseWrapper(apis);
const InstanceDispatch = vk.InstanceWrapper(apis);
const DeviceDispatch = vk.DeviceWrapper(apis);

const Instance = vk.InstanceProxy(apis);
const Device = vk.DeviceProxy(apis);
const Queue = vk.QueueProxy(apis);
const CommandBuffer = vk.CommandBufferProxy(apis);

pub const VulkanRenderer = struct {
    const Self = @This();

    allocator: std.mem.Allocator,

    vkb: BaseDispatch,

    window: sdl.Window,

    // Main
    instance: Instance,
    physical_device: vk.PhysicalDevice,
    device: Device,
    graphics_queue: Queue,
    presentation_queue: Queue,
    surface: vk.SurfaceKHR,
    swapchain: vk.SwapchainKHR,
    viewport: vk.Viewport,
    scissor: vk.Rect2D,

    swapchain_images: []SwapchainImage,
    swapchain_framebuffers: []vk.Framebuffer,
    command_buffers: []CommandBuffer,

    // Pipeline
    graphics_pipeline: vk.Pipeline,
    pipeline_layout: vk.PipelineLayout,
    render_pass: vk.RenderPass,

    // Pools
    graphics_command_pool: vk.CommandPool,

    // Utilities
    swapchain_image_format: vk.Format,
    extent: vk.Extent2D,

    // Synchronisation
    image_available: vk.Semaphore,
    render_finished: vk.Semaphore,
    frame_fence: vk.Fence,

    debug_utils: ?vk.DebugUtilsMessengerEXT,

    pub fn init(window: sdl.Window, allocator: std.mem.Allocator) !Self {
        var self: Self = undefined;

        self.window = window;
        self.allocator = allocator;
        self.vkb = try BaseDispatch.load(try sdl.vulkan.getVkGetInstanceProcAddr());

        try self.createInstance();
        try self.createSurface();

        if (enable_validation_layers) {
            self.debug_utils = try createDebugMessenger(self.instance);
        }

        try self.getPhysicalDevice();
        try self.createLogicalDevice();
        try self.createSwapchain();
        try self.createRenderPass();
        try self.createGraphicsPipeline();
        try self.createFramebuffers();
        try self.createCommandPool();
        try self.createCommandBuffers();
        try self.recordCommands();
        try self.createSynchronisation();

        return self;
    }

    pub fn draw(self: *Self) !void {
        _ = try self.device.waitForFences(1, @ptrCast(&self.frame_fence), vk.TRUE, std.math.maxInt(u64));
        _ = try self.device.resetFences(1, @ptrCast(&self.frame_fence));
        // -- Get next image
        // Get index of next image to be drawn to, and signal semaphore when ready to be drawn to
        const image_index_result = try self.device.acquireNextImageKHR(
            self.swapchain,
            std.math.maxInt(u64),
            self.image_available,
            .null_handle,
        );

        // -- Submit command buffer to render
        // Queue submission information
        const wait_stages = [_]vk.PipelineStageFlags{.{ .color_attachment_output_bit = true }};

        const submit_info: vk.SubmitInfo = .{
            .wait_semaphore_count = 1, // Number of semaphores to wait on
            .p_wait_semaphores = @ptrCast(&self.image_available), // List of semaphores to wait on
            .p_wait_dst_stage_mask = &wait_stages, // Stages to check semaphores at
            .command_buffer_count = 1, // Number of command buffers to submit
            .p_command_buffers = @ptrCast(&self.command_buffers[image_index_result.image_index]), // Command buffer to submit
            .signal_semaphore_count = 1, // Number of semaphores to signal
            .p_signal_semaphores = @ptrCast(&self.render_finished), // List of semaphores to signal when command buffer finishes
        };

        // Submit command buffer to queue
        try self.device.queueSubmit(self.graphics_queue.handle, 1, @ptrCast(&submit_info), self.frame_fence);

        // -- Present rendered image to screen
        const present_info: vk.PresentInfoKHR = .{
            .wait_semaphore_count = 1, // Number of semaphores to wait on
            .p_wait_semaphores = @ptrCast(&self.render_finished), // Semaphores to wait on
            .swapchain_count = 1, // Number of swapchains to present to
            .p_swapchains = @ptrCast(&self.swapchain), // Swapchains to present images to
            .p_image_indices = @ptrCast(&image_index_result.image_index), // Index of images in swapchains to present
        };

        // Present image
        _ = try self.device.queuePresentKHR(self.presentation_queue.handle, &present_info);
    }

    pub fn deinit(self: *Self) void {
        if (enable_validation_layers) {
            self.instance.destroyDebugUtilsMessengerEXT(self.debug_utils.?, null);
        }

        _ = self.device.waitForFences(1, @ptrCast(&self.frame_fence), vk.TRUE, std.math.maxInt(u64)) catch undefined;
        self.device.destroyFence(self.frame_fence, null);
        self.device.destroySemaphore(self.render_finished, null);
        self.device.destroySemaphore(self.image_available, null);

        self.allocator.free(self.command_buffers);
        self.device.destroyCommandPool(self.graphics_command_pool, null);

        for (self.swapchain_framebuffers) |framebuffer| {
            self.device.destroyFramebuffer(framebuffer, null);
        }

        self.allocator.free(self.swapchain_framebuffers);

        self.device.destroyPipeline(self.graphics_pipeline, null);
        self.device.destroyPipelineLayout(self.pipeline_layout, null);
        self.device.destroyRenderPass(self.render_pass, null);

        for (self.swapchain_images) |swapchain_image| {
            self.device.destroyImageView(swapchain_image.image_view, null);
        }

        self.allocator.free(self.swapchain_images);
        self.device.destroySwapchainKHR(self.swapchain, null);
        self.device.destroyDevice(null);
        self.instance.destroySurfaceKHR(self.surface, null);
        self.instance.destroyInstance(null);

        self.allocator.destroy(self.device.wrapper);
        self.allocator.destroy(self.instance.wrapper);
    }

    fn createInstance(self: *Self) !void {
        if (enable_validation_layers and !self.checkValidationLayersSupport()) {
            // TODO Better error
            return error.LayerNotPresent;
        }

        const extensions = try self.getRequiredExtensions();
        defer self.allocator.free(extensions);

        std.debug.print("[Required instance extensions]\n", .{});
        for (extensions) |ext| {
            std.debug.print("\t- {s}\n", .{ext});
        }

        if (!try self.checkInstanceExtensions(&extensions)) {
            return error.ExtensionNotPresent;
        }

        const app_info = vk.ApplicationInfo{
            .p_application_name = "Vulkan SDL Test",
            .application_version = vk.makeApiVersion(0, 0, 1, 0),
            .p_engine_name = "Vulkan SDL Test",
            .engine_version = vk.makeApiVersion(0, 0, 1, 0),
            .api_version = vk.API_VERSION_1_3,
        };

        var instance_create_info: vk.InstanceCreateInfo = .{
            .p_application_info = &app_info,
            .enabled_extension_count = @intCast(extensions.len),
            .pp_enabled_extension_names = @ptrCast(extensions),
        };

        if (enable_validation_layers) {
            const debug_create_info = getDebugUtilsCreateInfo();

            instance_create_info.enabled_layer_count = @intCast(validation_layers.len);
            instance_create_info.pp_enabled_layer_names = &validation_layers;
            instance_create_info.p_next = &debug_create_info;
        }

        const instance_handle = try self.vkb.createInstance(&instance_create_info, null);
        const vki = try self.allocator.create(InstanceDispatch);
        errdefer self.allocator.destroy(vki);
        vki.* = try InstanceDispatch.load(instance_handle, self.vkb.dispatch.vkGetInstanceProcAddr);

        self.instance = Instance.init(instance_handle, vki);
    }

    fn createSurface(self: *Self) !void {
        self.surface = try sdl.vulkan.createSurface(self.window, self.instance.handle);
    }

    fn createLogicalDevice(self: *Self) !void {
        const indices = try self.getQueueFamilies(self.physical_device);
        // 1 is the highest priority
        const priority = [_]f32{1};

        const qci = [_]vk.DeviceQueueCreateInfo{
            .{
                .queue_family_index = indices.graphics_family.?,
                .queue_count = 1,
                .p_queue_priorities = &priority,
            },
            .{
                .queue_family_index = indices.presentation_family.?,
                .queue_count = 1,
                .p_queue_priorities = &priority,
            },
        };

        const queue_count: u32 = if (indices.graphics_family.? == indices.presentation_family.?)
            1
        else
            2;

        const device_create_info: vk.DeviceCreateInfo = .{
            .queue_create_info_count = queue_count,
            .p_queue_create_infos = &qci,
            .pp_enabled_extension_names = &Utilities.device_extensions,
            .enabled_extension_count = @intCast(Utilities.device_extensions.len),
        };

        const device_handle = try self.instance.createDevice(self.physical_device, &device_create_info, null);

        const vkd = try self.allocator.create(DeviceDispatch);
        errdefer self.allocator.destroy(vkd);
        vkd.* = try DeviceDispatch.load(device_handle, self.instance.wrapper.dispatch.vkGetDeviceProcAddr);

        self.device = Device.init(device_handle, vkd);

        const queues = try self.getDeviceQueues();

        self.graphics_queue = Queue.init(queues[0], self.device.wrapper);
        self.presentation_queue = Queue.init(queues[1], self.device.wrapper);
    }

    fn createSwapchain(self: *Self) !void {
        const swapchain_details = try self.getSwapchainDetails(self.physical_device);
        defer self.allocator.free(swapchain_details.formats);
        defer self.allocator.free(swapchain_details.presentation_modes);

        // 1. Choose best surface format
        const surface_format = chooseBestSurfaceFormat(swapchain_details.formats);
        // 2. Choose best presentation mode
        const present_mode = chooseBestPresentationMode(swapchain_details.presentation_modes);
        // 3. Choose swapchain image resolution
        const extent = chooseSwapExtent(&self.window, swapchain_details.surface_capabilities);

        // How many images are in the swapchain? Get 1 more than the minimum to allow triple buffering
        var image_count: u32 = swapchain_details.surface_capabilities.min_image_count + 1;
        const max_image_count = swapchain_details.surface_capabilities.max_image_count;

        // Clamp down if higher
        // If 0, it means it's limitless
        if (max_image_count != 0 and image_count > max_image_count) {
            image_count = max_image_count;
        }

        var swapchain_create_info: vk.SwapchainCreateInfoKHR = .{
            .image_format = surface_format.format,
            .image_color_space = surface_format.color_space,
            .present_mode = present_mode,
            .image_extent = extent,
            .min_image_count = image_count,
            .image_array_layers = 1, // Number of layers for each image
            .image_usage = .{ .color_attachment_bit = true }, // What attachment will images be used as
            .pre_transform = swapchain_details.surface_capabilities.current_transform, // Transform to perform on swapchain images
            .composite_alpha = .{ .opaque_bit_khr = true }, // How to handle blending images with external graphics (e.g.: other windows)
            .clipped = vk.TRUE, // Whether to clip parts of images not in view (e.g.: behind another window, off-screen, etc...)
            .old_swapchain = .null_handle, // Links old one to quickly share responsibilities in case it's been destroyed and replaced
            .surface = self.surface,
            .image_sharing_mode = .exclusive,
        };

        // Get queue family indices
        const family_indices = try self.getQueueFamilies(self.physical_device);

        // If graphic and presentation families are different, then swapchain must let images be shared between families

        if (family_indices.graphics_family.? != family_indices.presentation_family.?) {
            const qfi = [_]u32{
                family_indices.graphics_family.?,
                family_indices.presentation_family.?,
            };

            swapchain_create_info.image_sharing_mode = .concurrent;
            swapchain_create_info.queue_family_index_count = 2; // Number of queues to share images between
            swapchain_create_info.p_queue_family_indices = &qfi;
        }

        self.swapchain = try self.device.createSwapchainKHR(&swapchain_create_info, null);
        self.swapchain_image_format = surface_format.format;
        self.extent = extent;

        // Swapchain images
        var swapchain_image_count: u32 = 0;
        _ = try self.device.getSwapchainImagesKHR(self.swapchain, &swapchain_image_count, null);

        const images = try self.allocator.alloc(vk.Image, swapchain_image_count);
        defer self.allocator.free(images);

        _ = try self.device.getSwapchainImagesKHR(self.swapchain, &swapchain_image_count, images.ptr);

        self.swapchain_images = try self.allocator.alloc(SwapchainImage, swapchain_image_count);

        for (images, 0..) |image, i| {
            self.swapchain_images[i] = .{
                .image = image,
                .image_view = try self.createImageView(image, self.swapchain_image_format, .{ .color_bit = true }),
            };
        }
    }

    fn createRenderPass(self: *Self) !void {
        // Colour attachment of the render pass
        const colour_attachment: vk.AttachmentDescription = .{
            .format = self.swapchain_image_format, // Format to use for attachment
            .samples = .{ .@"1_bit" = true }, // Number of samples to write for multisampling
            .load_op = .clear, // Describes what to do with attachment before rendering
            .store_op = .store, // Describes what to do with attachment after rendering
            .stencil_load_op = .dont_care, // Describes what to do with stencil before rendering
            .stencil_store_op = .dont_care, // Describes what to do with stencil after rendering
            // Framebuffer data will be stored as an image, but images can be given different data layouts
            // to give optimal use for certain operations
            .initial_layout = vk.ImageLayout.undefined, // Image data layout before render pass starts
            .final_layout = vk.ImageLayout.present_src_khr, // Image data layout after render pass (to change to)
        };

        // Attachment reference uses an attachment index that refers to index in the attachment list passed to render pass create info
        const colour_attachment_reference: vk.AttachmentReference = .{
            .attachment = 0,
            .layout = vk.ImageLayout.color_attachment_optimal,
        };

        // Information about a particular subpass the render pass is using
        const subpass: vk.SubpassDescription = .{
            .pipeline_bind_point = .graphics, // Pipeline type subpass is to be bound to
            .color_attachment_count = 1,
            .p_color_attachments = @ptrCast(&colour_attachment_reference),
        };

        // Need to determine when layout transitions occur using subpass dependencies
        const subpass_dependencies = [2]vk.SubpassDependency{
            // Conversion from VK_IMAGE_LAYOUT_UNDEFINED to VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL
            vk.SubpassDependency{
                // Transition must happen after...
                .src_subpass = vk.SUBPASS_EXTERNAL, // Subpass index (VK_SUBPASS_EXTERNAL = outside of renderpass)
                .src_stage_mask = .{ .bottom_of_pipe_bit = true }, // Pipeline stage
                .src_access_mask = .{ .memory_read_bit = true }, // Stage access mask (memory access)
                // But must happen before...
                .dst_subpass = 0,
                .dst_stage_mask = .{ .color_attachment_output_bit = true },
                .dst_access_mask = .{ .memory_read_bit = true, .memory_write_bit = true },
            },
            // Conversion from VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL to VK_IMAGE_LAYOUT_PRESENT_SRC_KHR
            vk.SubpassDependency{
                // Transition must happen after...
                .src_subpass = 0,
                .src_stage_mask = .{ .color_attachment_output_bit = true },
                .src_access_mask = .{ .memory_read_bit = true, .memory_write_bit = true },
                // But must happen before...
                .dst_subpass = vk.SUBPASS_EXTERNAL,
                .dst_stage_mask = .{ .bottom_of_pipe_bit = true },
                .dst_access_mask = .{ .memory_read_bit = true },
            },
        };

        const render_pass_create_info: vk.RenderPassCreateInfo = .{
            .attachment_count = 1,
            .p_attachments = @ptrCast(&colour_attachment),
            .subpass_count = 1,
            .p_subpasses = @ptrCast(&subpass),
            .dependency_count = @intCast(subpass_dependencies.len),
            .p_dependencies = &subpass_dependencies,
        };

        self.render_pass = try self.device.createRenderPass(&render_pass_create_info, null);
    }

    fn createGraphicsPipeline(self: *Self) !void {
        // Create shader modules
        const vert = try self.device.createShaderModule(&.{
            .code_size = shaders.shader_vert.len,
            .p_code = @ptrCast(&shaders.shader_vert),
        }, null);
        defer self.device.destroyShaderModule(vert, null);

        const frag = try self.device.createShaderModule(&.{
            .code_size = shaders.shader_frag.len,
            .p_code = @ptrCast(&shaders.shader_frag),
        }, null);
        defer self.device.destroyShaderModule(frag, null);

        // -- Shader stage creation information --

        // Vertex stage creation information
        const vertex_shader_create_info: vk.PipelineShaderStageCreateInfo = .{
            .stage = .{ .vertex_bit = true },
            .module = vert,
            .p_name = "main",
        };

        // Fragment stage creation information
        const fragment_shader_create_info: vk.PipelineShaderStageCreateInfo = .{
            .stage = .{ .fragment_bit = true },
            .module = frag,
            .p_name = "main",
        };

        const shader_create_infos = [_]vk.PipelineShaderStageCreateInfo{
            vertex_shader_create_info,
            fragment_shader_create_info,
        };

        // -- Vertex input (TODO: Put in vertex descriptions when resources created) --
        const vertex_input_create_info: vk.PipelineVertexInputStateCreateInfo = .{
            .p_vertex_binding_descriptions = null, // List of vertex binding descriptions (data spacing, stride info)
            .p_vertex_attribute_descriptions = null, // List of vertex attribute descriptions (data format and where to bind to/from)
        };

        // -- Input assembly --
        const assembly_create_info: vk.PipelineInputAssemblyStateCreateInfo = .{
            .topology = .triangle_list, // Primitive type to assemble vertices as
            .primitive_restart_enable = vk.FALSE, // Allow overrinding of strip topology to start new primitives
        };

        // -- Viewport & scissor --
        self.viewport = .{
            .x = 0.0,
            .y = 0.0,
            .width = @floatFromInt(self.extent.width),
            .height = @floatFromInt(self.extent.height),
            .min_depth = 0.0,
            .max_depth = 1.0,
        };

        self.scissor = .{
            .offset = .{ .x = 0, .y = 0 },
            .extent = self.extent,
        };

        const viewport_state_create_info: vk.PipelineViewportStateCreateInfo = .{
            .viewport_count = 1,
            .p_viewports = @ptrCast(&self.viewport),
            .scissor_count = 1,
            .p_scissors = @ptrCast(&self.scissor),
        };

        // -- Dynamic states --
        // Dynamic states to enable (TODO: To investigate later)
        const dynamic_states = [_]vk.DynamicState{ .viewport, .scissor };

        const dynamic_state_create_info: vk.PipelineDynamicStateCreateInfo = .{
            .dynamic_state_count = 2,
            .p_dynamic_states = &dynamic_states,
        };

        // -- Rasterizer --

        const rasterizer_create_info: vk.PipelineRasterizationStateCreateInfo = .{
            .depth_clamp_enable = vk.FALSE, // Change if fragments beyond near/far planes are clipped (default) or clamped to planed
            .rasterizer_discard_enable = vk.FALSE, // Whether to discard data and skip rasterizer (never creates fragments)
            .polygon_mode = .fill, // How to handle filling points between vertices
            .line_width = 1.0, // How thick the lines should be when drawn
            .cull_mode = .{ .back_bit = true }, // Which face of a triangle to cull
            .front_face = .clockwise, // Winding to determine which side is front
            .depth_bias_enable = vk.FALSE, // Whether to add depth bias to fragments (good for stopping "shadow acne" in shadow mapping)
            .depth_bias_constant_factor = 0,
            .depth_bias_clamp = 0,
            .depth_bias_slope_factor = 0,
        };

        // -- Multisampling --
        const multisampling_create_info: vk.PipelineMultisampleStateCreateInfo = .{
            .sample_shading_enable = vk.FALSE, // Enable multisample shading or not
            .rasterization_samples = .{ .@"1_bit" = true }, // Number of samples to use per fragment
            .min_sample_shading = 1,
            .alpha_to_coverage_enable = vk.FALSE,
            .alpha_to_one_enable = vk.FALSE,
        };

        // -- Blending --

        // Blend attachment state (how blending is handled)
        const colour_state: vk.PipelineColorBlendAttachmentState = .{
            .color_write_mask = .{ .r_bit = true, .g_bit = true, .b_bit = true, .a_bit = true }, // Colours to apply blending to
            .blend_enable = vk.TRUE, // Enable blending
            .src_color_blend_factor = vk.BlendFactor.src_alpha,
            .dst_color_blend_factor = vk.BlendFactor.one_minus_src_alpha,
            .color_blend_op = vk.BlendOp.add,
            // Summary: (VK_BLEND_FACTOR_SRC_ALPHA * new colour) + (VK_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA * old colour)
            .src_alpha_blend_factor = vk.BlendFactor.one,
            .dst_alpha_blend_factor = vk.BlendFactor.zero,
            .alpha_blend_op = vk.BlendOp.add,
            // Summary (1 * new alpha) + (0 * old alpha) = new alpha
        };

        // Blending uses equation: (srcColorBlendFactor * new colour) colorBlendOp (dstColorBlendFactor * old colour)

        const colour_blending_create_info: vk.PipelineColorBlendStateCreateInfo = .{
            .logic_op_enable = vk.FALSE, // Alternative to calculations is to use logical operations
            .logic_op = .copy,
            .attachment_count = 1,
            .p_attachments = @ptrCast(&colour_state),
            .blend_constants = [_]f32{ 0, 0, 0, 0 },
        };

        // -- Pipeline layout (TODO: Apply future descriptor set layouts) --
        const pipeline_layout_create_info: vk.PipelineLayoutCreateInfo = .{};

        self.pipeline_layout = try self.device.createPipelineLayout(&pipeline_layout_create_info, null);

        // -- Depth stencil testing --
        // TODO: Set a depth stencil testing

        // -- Graphics pipeline creation --
        const pipeline_create_info: vk.GraphicsPipelineCreateInfo = .{
            .stage_count = 2, // Number of shader stages
            .p_stages = &shader_create_infos, // List of shader stages
            .p_vertex_input_state = &vertex_input_create_info,
            .p_input_assembly_state = &assembly_create_info,
            .p_viewport_state = &viewport_state_create_info,
            .p_dynamic_state = &dynamic_state_create_info,
            .p_rasterization_state = &rasterizer_create_info,
            .p_multisample_state = &multisampling_create_info,
            .p_color_blend_state = &colour_blending_create_info,
            .p_depth_stencil_state = null,
            .layout = self.pipeline_layout, // Pipeline layout the pipeline should use
            .render_pass = self.render_pass, // Renderpass description the pipeline is compatible with
            .subpass = 0, // Subpass of renderpass to use with pipeline
            // Pipeline derivatives: can create multiple pipelines that derive from one another for optimisation
            .base_pipeline_handle = .null_handle, // Existing pipeline to derive from...
            .base_pipeline_index = -1, // Or index of pipeline being created to derive from (in case creating multiple at once)
        };

        _ = try self.device.createGraphicsPipelines(
            .null_handle,
            1,
            @ptrCast(&pipeline_create_info),
            null,
            @ptrCast(&self.graphics_pipeline),
        );
    }

    fn createFramebuffers(self: *Self) !void {
        self.swapchain_framebuffers = try self.allocator.alloc(vk.Framebuffer, self.swapchain_images.len);

        // Create a frammebuffer for each swapchain image
        for (self.swapchain_images, 0..) |swapchain_image, i| {
            const attachments = [_]vk.ImageView{swapchain_image.image_view};

            const framebuffer_create_info: vk.FramebufferCreateInfo = .{
                .render_pass = self.render_pass, // Render pass layout the frambuffer will be used with
                .attachment_count = @intCast(attachments.len),
                .p_attachments = &attachments, // List of attachments (1:1 with render pass)
                .width = self.extent.width, // Framebuffer width
                .height = self.extent.height, // Framebuffer height
                .layers = 1, // Framebuffer layers
            };

            self.swapchain_framebuffers[i] = try self.device.createFramebuffer(&framebuffer_create_info, null);
        }
    }

    fn createCommandPool(self: *Self) !void {
        // Get indices of queue families from device
        const queue_family_indices = try self.getQueueFamilies(self.physical_device);

        const pool_create_info: vk.CommandPoolCreateInfo = .{
            // Queue family type that buffers from this command pool will use
            .queue_family_index = queue_family_indices.graphics_family.?,
        };

        // Create a graphics queue family command pool
        self.graphics_command_pool = try self.device.createCommandPool(&pool_create_info, null);
    }

    fn createCommandBuffers(self: *Self) !void {
        // Allocate one command buffer for each framebuffer
        const command_buffer_handles = try self.allocator.alloc(vk.CommandBuffer, self.swapchain_framebuffers.len);
        defer self.allocator.free(command_buffer_handles);
        self.command_buffers = try self.allocator.alloc(CommandBuffer, command_buffer_handles.len);

        const command_buffer_allocate_info: vk.CommandBufferAllocateInfo = .{
            .command_pool = self.graphics_command_pool,
            .level = .primary, // primary: buffer you submit directly to queue. Can't be called by other buffers
            .command_buffer_count = @intCast(command_buffer_handles.len),
        };

        // Allocate command buffers and place handles in array of buffers
        try self.device.allocateCommandBuffers(&command_buffer_allocate_info, command_buffer_handles.ptr);
        for (command_buffer_handles, 0..) |command_buffer_handle, i| {
            self.command_buffers[i] = CommandBuffer.init(command_buffer_handle, self.device.wrapper);
        }
    }

    fn createSynchronisation(self: *Self) !void {
        // Semaphore creation information
        self.image_available = try self.device.createSemaphore(&.{}, null);
        self.render_finished = try self.device.createSemaphore(&.{}, null);

        const fence_create_info: vk.FenceCreateInfo = .{ .flags = .{ .signaled_bit = true } };
        self.frame_fence = try self.device.createFence(&fence_create_info, null);
    }

    fn recordCommands(self: *Self) !void {
        // Information about how to begin each command
        const buffer_begin_info: vk.CommandBufferBeginInfo = .{
            // Buffer can be resubmitted when it has already been submitted and is awaiting execution
            .flags = .{ .simultaneous_use_bit = true },
        };

        const clear_values = [_]vk.ClearValue{
            .{ .color = .{ .float_32 = [4]f32{ 0.6, 0.65, 0.4, 1.0 } } },
        };

        // Information about how to begin a render pass (only needed for graphical application)
        var render_pass_begin_info: vk.RenderPassBeginInfo = .{
            .render_pass = self.render_pass, // Render pass to begin
            .render_area = .{
                .offset = .{ .x = 0, .y = 0 }, // Start point of render pass in pixels
                .extent = self.extent, // Size of region to run render pass on (starting at offset)
            },
            .p_clear_values = &clear_values, // List of clear values (TODO: Depth attachment clear value)
            .clear_value_count = 1,
            .framebuffer = undefined,
        };

        for (self.command_buffers, 0..) |command_buffer, i| {
            render_pass_begin_info.framebuffer = self.swapchain_framebuffers[i];

            // Start recording commands to command buffer
            try command_buffer.beginCommandBuffer(&buffer_begin_info);

            {
                // Begin render pass
                command_buffer.beginRenderPass(&render_pass_begin_info, vk.SubpassContents.@"inline");

                // Bind pipeline to be used in render pass
                command_buffer.bindPipeline(.graphics, self.graphics_pipeline);

                // Needed when using dynamic state
                command_buffer.setViewport(0, 1, @ptrCast(&self.viewport));
                command_buffer.setScissor(0, 1, @ptrCast(&self.scissor));

                // Execute a pipeline
                command_buffer.draw(3, 1, 0, 0);

                // End render pass
                command_buffer.endRenderPass();
            }

            // Stop recording to command buffer
            try command_buffer.endCommandBuffer();
        }
    }

    fn getPhysicalDevice(self: *Self) !void {
        var pdev_count: u32 = 0;
        _ = try self.instance.enumeratePhysicalDevices(&pdev_count, null);

        const pdevs = try self.allocator.alloc(vk.PhysicalDevice, pdev_count);
        defer self.allocator.free(pdevs);

        _ = try self.instance.enumeratePhysicalDevices(&pdev_count, pdevs.ptr);

        for (pdevs) |pdev| {
            if (self.checkDeviceSuitable(pdev)) {
                self.physical_device = pdev;
                break;
            }
        } else {
            // TODO Obviously needs to be something else
            unreachable;
        }
    }

    fn getRequiredExtensions(self: Self) ![][*:0]const u8 {
        var ext_count = sdl.vulkan.getInstanceExtensionsCount(self.window);

        if (enable_validation_layers) {
            ext_count += 1;
        }

        var extensions = try self.allocator.alloc([*:0]const u8, ext_count);
        _ = try sdl.vulkan.getInstanceExtensions(self.window, extensions);

        if (enable_validation_layers) {
            extensions[extensions.len - 1] = vk.extensions.ext_debug_utils.name;
        }

        return extensions;
    }

    fn getQueueFamilies(self: Self, pdev: vk.PhysicalDevice) !QueueFamilyIndices {
        var indices: QueueFamilyIndices = .{ .graphics_family = null };

        var queue_family_count: u32 = 0;
        self.instance.getPhysicalDeviceQueueFamilyProperties(pdev, &queue_family_count, null);

        const queue_family_list = try self.allocator.alloc(vk.QueueFamilyProperties, queue_family_count);
        defer self.allocator.free(queue_family_list);

        self.instance.getPhysicalDeviceQueueFamilyProperties(pdev, &queue_family_count, queue_family_list.ptr);

        for (queue_family_list, 0..) |queue_family, i| {
            if (queue_family.queue_count > 0 and queue_family.queue_flags.graphics_bit) {
                indices.graphics_family = @intCast(i);
            }

            const presentation_support = try self.instance.getPhysicalDeviceSurfaceSupportKHR(pdev, @intCast(i), self.surface);
            if (queue_family.queue_count > 0 and presentation_support == vk.TRUE) {
                indices.presentation_family = @intCast(i);
            }

            if (indices.isValid()) {
                return indices;
            }
        }

        unreachable;
    }

    fn getDeviceQueues(self: Self) ![2]vk.Queue {
        const indices = try self.getQueueFamilies(self.physical_device);

        const graphics_queue = self.device.getDeviceQueue(indices.graphics_family.?, 0);
        const presentation_queue = self.device.getDeviceQueue(indices.presentation_family.?, 0);
        return .{ graphics_queue, presentation_queue };
    }

    fn checkInstanceExtensions(self: Self, required_extensions: *const [][*:0]const u8) !bool {
        var prop_count: u32 = 0;
        _ = try self.vkb.enumerateInstanceExtensionProperties(null, &prop_count, null);

        const props = try self.allocator.alloc(vk.ExtensionProperties, prop_count);
        defer self.allocator.free(props);

        _ = try self.vkb.enumerateInstanceExtensionProperties(null, &prop_count, props.ptr);

        for (required_extensions.*) |required_extension| {
            for (props) |prop| {
                if (std.mem.eql(u8, std.mem.sliceTo(&prop.extension_name, 0), std.mem.span(required_extension))) {
                    break;
                }
            } else {
                return false;
            }
        }

        return true;
    }

    fn checkDeviceExtensions(self: Self, pdev: vk.PhysicalDevice) !bool {
        var prop_count: u32 = 0;
        _ = try self.instance.enumerateDeviceExtensionProperties(pdev, null, &prop_count, null);

        if (prop_count == 0) {
            return false;
        }

        const props = try self.allocator.alloc(vk.ExtensionProperties, prop_count);
        defer self.allocator.free(props);

        _ = try self.instance.enumerateDeviceExtensionProperties(pdev, null, &prop_count, props.ptr);

        for (Utilities.device_extensions) |device_extension| {
            for (props) |prop| {
                if (std.mem.eql(u8, std.mem.sliceTo(&prop.extension_name, 0), std.mem.span(device_extension))) {
                    break;
                }
            } else {
                return false;
            }
        }

        return true;
    }

    fn checkDeviceSuitable(self: Self, pdev: vk.PhysicalDevice) bool {
        const pdev_properties = self.instance.getPhysicalDeviceProperties(pdev);

        if (pdev_properties.device_type == .cpu) {
            return false;
        }

        const queue_family_indices = self.getQueueFamilies(pdev) catch return false;

        const extension_support = self.checkDeviceExtensions(pdev) catch return false;

        const swapchain_details = self.getSwapchainDetails(pdev) catch return false;
        defer self.allocator.free(swapchain_details.formats);
        defer self.allocator.free(swapchain_details.presentation_modes);

        const swapchain_valid = swapchain_details.formats.len != 0 and swapchain_details.formats.len != 0;

        return queue_family_indices.isValid() and extension_support and swapchain_valid;
    }

    fn checkValidationLayersSupport(self: Self) bool {
        var layer_count: u32 = undefined;
        _ = self.vkb.enumerateInstanceLayerProperties(&layer_count, null) catch return false;

        const available_layers = self.allocator.alloc(vk.LayerProperties, layer_count) catch unreachable;
        defer self.allocator.free(available_layers);

        _ = self.vkb.enumerateInstanceLayerProperties(&layer_count, available_layers.ptr) catch return false;

        for (validation_layers) |validation_layer| {
            for (available_layers) |available_layer| {
                if (std.mem.eql(u8, std.mem.span(validation_layer), std.mem.sliceTo(&available_layer.layer_name, 0))) {
                    return true;
                }
            }
        }

        return false;
    }

    fn getSwapchainDetails(self: Self, pdev: vk.PhysicalDevice) !SwapchainDetails {
        // Capabilities
        const surface_capabilities = try self.instance.getPhysicalDeviceSurfaceCapabilitiesKHR(pdev, self.surface);

        // Formats
        var format_count: u32 = 0;
        _ = try self.instance.getPhysicalDeviceSurfaceFormatsKHR(pdev, self.surface, &format_count, null);

        const formats = try self.allocator.alloc(vk.SurfaceFormatKHR, format_count);
        _ = try self.instance.getPhysicalDeviceSurfaceFormatsKHR(pdev, self.surface, &format_count, formats.ptr);

        // Presentation modes
        var present_mode_count: u32 = 0;
        _ = try self.instance.getPhysicalDeviceSurfacePresentModesKHR(pdev, self.surface, &present_mode_count, null);

        const presentation_modes = try self.allocator.alloc(vk.PresentModeKHR, format_count);
        _ = try self.instance.getPhysicalDeviceSurfacePresentModesKHR(pdev, self.surface, &present_mode_count, presentation_modes.ptr);

        return .{
            .surface_capabilities = surface_capabilities,
            .formats = formats,
            .presentation_modes = presentation_modes,
        };
    }

    fn createImageView(self: Self, image: vk.Image, format: vk.Format, aspect_flags: vk.ImageAspectFlags) !vk.ImageView {
        const image_view_create_info: vk.ImageViewCreateInfo = .{
            .image = image,
            .format = format,
            .view_type = .@"2d",
            .components = .{
                // Used for remapping rgba values to other rgba values
                .r = .identity,
                .g = .identity,
                .b = .identity,
                .a = .identity,
            },
            .subresource_range = .{
                .aspect_mask = aspect_flags, // Which aspect of image to view (e.g.: colour, depth, stencil, etc...)
                .base_mip_level = 0, // Start mipmap level to view from
                .level_count = 1, // Number of mipmap levels to view
                .base_array_layer = 0, // Start array level to view from
                .layer_count = 1, // Number of array levels to view
            },
        };

        return try self.device.createImageView(&image_view_create_info, null);
    }
};

// Format: VK_FORMAT_R8G8B8A8_UNORM (VK_FORMAT_B8G8R8A8_UNORM as backup)
// Color space: VK_COLOR_SPACE_SRGB_NONLINEAR_KHR
fn chooseBestSurfaceFormat(formats: []vk.SurfaceFormatKHR) vk.SurfaceFormatKHR {
    // If only one format available and is undefined, then this means all formats are available
    if (formats.len == 1 and formats[0].format == vk.Format.undefined) {
        return .{
            .format = vk.Format.r8g8b8a8_unorm,
            .color_space = vk.ColorSpaceKHR.srgb_nonlinear_khr,
        };
    }

    for (formats) |format| {
        if ((format.format == vk.Format.r8g8b8a8_unorm or format.format == vk.Format.b8g8r8a8_unorm) and format.color_space == vk.ColorSpaceKHR.srgb_nonlinear_khr) {
            return format;
        }
    }

    return formats[0];
}

fn chooseBestPresentationMode(presentation_modes: []vk.PresentModeKHR) vk.PresentModeKHR {
    for (presentation_modes) |presentation_mode| {
        if (presentation_mode == vk.PresentModeKHR.mailbox_khr) {
            return presentation_mode;
        }
    }

    // Use FIFO as Vulkan spec says it must be present
    return vk.PresentModeKHR.fifo_khr;
}

fn chooseSwapExtent(window: *sdl.Window, surface_capabilities: vk.SurfaceCapabilitiesKHR) vk.Extent2D {
    // If the current extent is at max value, the extent can vary. Otherwise it's the size of the window
    if (surface_capabilities.current_extent.width != std.math.maxInt(u32)) {
        return surface_capabilities.current_extent;
    }

    // If value can very, need to set the extent manually
    const framebuffer_size = sdl.vulkan.getDrawableSize(window);

    var extent: vk.Extent2D = .{
        .width = @intCast(framebuffer_size.width),
        .height = @intCast(framebuffer_size.height),
    };

    // Surface also defines max and min, so make sure it's within boundaries by clamping values
    extent.width = @max(surface_capabilities.min_image_extent.width, @min(surface_capabilities.max_image_extent.width, extent.width));
    extent.height = @max(surface_capabilities.min_image_extent.height, @min(surface_capabilities.max_image_extent.height, extent.height));

    return extent;
}

// Validation layers stuff
fn createDebugMessenger(instance: Instance) !vk.DebugUtilsMessengerEXT {
    const debug_create_info = getDebugUtilsCreateInfo();

    return try instance.createDebugUtilsMessengerEXT(&debug_create_info, null);
}

fn getDebugUtilsCreateInfo() vk.DebugUtilsMessengerCreateInfoEXT {
    return vk.DebugUtilsMessengerCreateInfoEXT{
        .message_severity = .{ .verbose_bit_ext = true, .warning_bit_ext = true, .error_bit_ext = true },
        .message_type = .{ .general_bit_ext = true, .validation_bit_ext = true, .performance_bit_ext = true },
        .pfn_user_callback = debugCallback,
    };
}

// message_severity: vk.DebugUtilsMessageSeverityFlagsEXT,
// message_types: vk.DebugUtilsMessageTypeFlagsEXT,
// p_callback_data: ?*const vk.DebugUtilsMessengerCallbackDataEXT,
// p_user_data: ?*anyopaque,
fn debugCallback(
    message_severity: vk.DebugUtilsMessageSeverityFlagsEXT,
    message_types: vk.DebugUtilsMessageTypeFlagsEXT,
    p_callback_data: ?*const vk.DebugUtilsMessengerCallbackDataEXT,
    _: ?*anyopaque,
) callconv(vk.vulkan_call_conv) vk.Bool32 {
    const severity = getMessageSeverityLabel(message_severity);
    const message_type = getMessageTypeLabel(message_types);

    std.debug.print("[{s}] ({s}): {s}\n", .{ severity, message_type, p_callback_data.?.p_message.? });
    return vk.TRUE;
}

inline fn getMessageSeverityLabel(message_severity: vk.DebugUtilsMessageSeverityFlagsEXT) []const u8 {
    if (message_severity.verbose_bit_ext) {
        return "VERBOSE";
    } else if (message_severity.info_bit_ext) {
        return "INFO";
    } else if (message_severity.warning_bit_ext) {
        return "WARNING";
    } else if (message_severity.error_bit_ext) {
        return "ERROR";
    } else {
        unreachable;
    }
}

inline fn getMessageTypeLabel(message_types: vk.DebugUtilsMessageTypeFlagsEXT) []const u8 {
    if (message_types.general_bit_ext) {
        return "general";
    } else if (message_types.validation_bit_ext) {
        return "validation";
    } else if (message_types.performance_bit_ext) {
        return "performance";
    } else if (message_types.device_address_binding_bit_ext) {
        return "device_address_binding";
    } else {
        return "unknown";
    }
}
